-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.sha3_jh_package.all;

entity jh_control_mem_ppl is		
	generic ( 		  
		PPL : integer := 2;
		HS : integer := 256;
		UF : integer := 1
	);
	port (					
		rst			: in std_logic;
		clk			: in std_logic;
		
		-- datapath signals
			--fsm1
		ein, ec, lc			: out std_logic_vector(ppl-1 downto 0);
		zc0, final_segment	: in std_logic_vector(ppl-1 downto 0);	 
		load_id : out std_logic_vector(ppl-1 downto 0);
        
        -- pad
        comp_rem_e0, comp_lb_e0 : in std_logic_vector(ppl-1 downto 0);	 
        en_lb, clr_lb           : out std_logic_vector(ppl-1 downto 0);	 
        spos, sel_pad           : out std_logic_vector(PPL*2-1 downto 0);
		en_size, rst_size       : out std_logic_vector(ppl-1 downto 0);	 
		
		--fsm2	   
		smsg : out std_logic_vector(log2(ppl)-1 downto 0);
		er, sf  :  out std_logic;
		schain  : out std_logic;
		round : out std_logic_vector(5-(UF-1) downto 0);
		emsg, load_hash : out std_logic_vector(ppl-1 downto 0);
			-- FSM3
		shift_dout		: out std_logic_vector(ppl-1 downto 0);
		sout 		: out std_logic_vector(log2(ppl)-1 downto 0);
		sid : out std_logic;
			--top fsm			
		
		-- fifo signals
		src_ready	: in std_logic_vector(ppl-1 downto 0);
		src_read	: out std_logic_vector(ppl-1 downto 0);
		dst_ready 	: in std_logic;
		dst_write	: out std_logic
	);				 
end jh_control_mem_ppl;

architecture struct of jh_control_mem_ppl is				   	 
	-- fsm1
	signal block_ready_set, msg_end_set, load_next_block : std_logic_vector(ppl-1 downto 0);	
	-- fsm2						 
		-- fsm1 communications
	signal block_ready_clr, msg_end_clr : std_logic_vector(ppl-1 downto 0); --out
	signal block_ready, msg_end : std_logic_vector(ppl-1 downto 0); --in	   
	
		-- fsm2 communications
	signal output_busy_set : std_logic_vector(ppl-1 downto 0); --out
	signal output_busy : std_logic_vector(ppl-1 downto 0); --in  
	
	-- fsm3										
	signal output_busy_clr : std_logic_vector(ppl-1 downto 0); 
	signal shift_dout_fsm3, load_hash_fsm2 : std_logic_vector(ppl-1 downto 0);
	
	-- pipelined signals
	signal er_sig, sf_sig, schain_sig, dst_write_sig, sid_sig  : std_logic;
	signal smsg_sig : std_logic_vector(log2(ppl)-1 downto 0);
	signal emsg_sig : std_logic_vector( ppl-1 downto 0);
	signal sout_sig : std_logic_vector(log2(ppl)-1 downto 0);
	
begin
	
	gfsm1: for i in 0 to ppl-1 generate
		fsm1_gen : entity work.jh_fsm1(beh) port map (
			clk => clk, rst => rst, 
			ein => ein(i), ec => ec(i), lc => lc(i), zc0 => zc0(i), final_segment => final_segment(i),
			load_next_block => load_next_block(i), block_ready_set => block_ready_set(i), msg_end_set => msg_end_set(i),
			src_ready => src_ready(i), src_read => src_read(i), load_id => load_id(i),
            -- pad
            comp_rem_e0 => comp_rem_e0(i), comp_lb_e0 => comp_lb_e0(i), en_lb => en_lb(i), clr_lb => clr_lb(i),
            spos => spos((PPL-i)*2-1 downto (PPL-i-1)*2), 
            sel_pad => sel_pad((PPL-i)*2-1 downto (PPL-i-1)*2), 
            en_size => en_size(i), rst_size => rst_size(i)
		);	 						
	end generate;
	
	fsm2_gen : entity work.jh_fsm2_mem(beh) 
	generic map ( ppl=> ppl, UF => UF )
	port map (
		clk => clk, rst => rst, 
		er => er_sig, emsg => emsg_sig , load_hash => load_hash_fsm2, sf => sf_sig, schain => schain_sig, smsg => smsg_sig,
		block_ready_clr => block_ready_clr, msg_end_clr => msg_end_clr, block_ready => block_ready, msg_end => msg_end,	round => round,
		output_busy_set => output_busy_set, output_busy => output_busy
	); 
	
	fsm3_gen : entity work.sha3_fsm3_ppl(beh)
		generic map ( h => HS, w=> w, ppl => ppl )
		port map (
		clk => clk, rst => rst, 
		shift_dout => shift_dout_fsm3, sout => sout_sig,  sid => sid_sig,
		output_busy => output_busy, output_busy_clr => output_busy_clr,
		dst_ready => dst_ready, dst_write => dst_write_sig
	);	 											   
	
	process ( clk )
	begin
		if rising_edge( clk ) then
			er <= er_sig;
			emsg <= emsg_sig;
			sf <= sf_sig;
			schain <= schain_sig;
			smsg <= smsg_sig;
			sout <= sout_sig;
			sid <= sid_sig;			
			shift_dout <= shift_dout_fsm3 or load_hash_fsm2; 
			load_hash  <= load_hash_fsm2;	
			dst_write <= dst_write_sig;
		end if;
	end process;
	
	comm :  for i in 0 to ppl-1 generate
		load_next_block(i) <= not block_ready(i);
		sr_blk_ready : sr_reg 
			port map ( rst => rst, clk => clk, set => block_ready_set(i), clr => block_ready_clr(i), output => block_ready(i));	
		sr_msg_end : sr_reg 
			port map ( rst => rst, clk => clk, set => msg_end_set(i), clr => msg_end_clr(i), output => msg_end(i));	
		sr_output_write : sr_reg 
			port map ( rst => rst, clk => clk, set => output_busy_set(i), clr => output_busy_clr(i), output => output_busy(i));			
	end generate; 
end struct;					