-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.sha3_jh_package.all;
use work.sha3_pkg.all;

entity jh_datapath_mem_ppl is	
	generic ( ppl : integer := 2; HS : integer := 256; UF : integer := 1 );
	port (
		-- external
		rst     : in std_logic;
		clk     : in std_logic;
		din     : in std_logic_vector(ppl*w-1 downto 0);	
		dout    : out std_logic_vector(w-1 downto 0);	
		
        --fsm 1		
		ein     : in std_logic_vector(ppl-1 downto 0);	
        ec      : in std_logic_vector(ppl-1 downto 0);	
        lc      : in std_logic_vector(ppl-1 downto 0);	
		zc0     : out std_logic_vector(ppl-1 downto 0);		   
		-- pad
        spos    : in std_logic_vector(PPL*2-1 downto 0);
		sel_pad : in std_logic_vector(PPL*2-1 downto 0);
		rst_size: in std_logic_vector(ppl-1 downto 0);	
        en_size : in std_logic_vector(ppl-1 downto 0);	
        en_lb   : in std_logic_vector(ppl-1 downto 0);	
        clr_lb  : in std_logic_vector(ppl-1 downto 0);	
        comp_rem_e0: out std_logic_vector(ppl-1 downto 0);
        comp_lb_e0 : out std_logic_vector(ppl-1 downto 0);

		--Round		  		
        emsg        : in std_logic_vector(ppl-1 downto 0);
		schain      : in std_logic;
		round       : in std_logic_vector(5-(UF-1) downto 0);
		smsg        : in std_logic_vector(log2(ppl)-1 downto 0);
		er          : in std_logic;
		sf          : in std_logic;
		
		-- Output				
		shift_dout  : in std_logic_vector(ppl-1 downto 0);
        load_id     : in std_logic_vector(ppl-1 downto 0);
        load_hash   : in std_logic_vector(ppl-1 downto 0);
		sout        : in std_logic_vector(log2(ppl)-1 downto 0);
		sid         : in std_logic
	);				  
end jh_datapath_mem_ppl;


architecture struct of jh_datapath_mem_ppl is 
	-- type declaration
	type block_array_type is array (0 to ppl-1) of std_logic_vector(b-1 downto 0);
	type mw_array_type is array ( 0 to ppl-1 ) of std_logic_vector(mw-1 downto 0);
	type c_array_type is array ( 0 to ppl-1) of std_logic_vector(15 downto 0);

	-- Padding related signals
	constant zeros : std_logic_vector(HS-1 downto 0) := (others => '0');
	
	type rem_bit_type    is array (0 to PPL-1) of std_logic_vector(8 downto 0);
	type ppl_8bit_type   is array (0 to PPL-1) of std_logic_vector(7 downto 0);
	type ppl_wbit_type   is array (0 to PPL-1) of std_logic_vector(63 downto 0);
	type ppl_sel2_type   is array (0 to PPL-1) of std_logic_vector(1 downto 0);
	signal remaining_bits           : rem_bit_type;
	signal din_padded               : ppl_wbit_type;
	signal din_or_length            : ppl_wbit_type;
	signal sel_pad_location_lookup  : ppl_8bit_type;
    signal sel_din_lookup           : ppl_8bit_type;
	signal sel_din                  : ppl_8bit_type; 
    signal sel_pad_location         : ppl_8bit_type; 
	signal sel_pad_2bits            : ppl_sel2_type;
    
	--input		  
	signal min, mreg : mw_array_type;
	constant iv : std_logic_vector(b-1 downto 0) := get_iv( HS );
	-- round constant
	type crd_out_ux_type is array (0 to UF-1) of std_logic_vector(crw-1 downto 0);
	signal crd_out, crd_out_pre : crd_out_ux_type;
	
	-- round		
	type rd_out_ux_type is array (0 to UF-1) of std_logic_vector(b-1 downto 0);
	signal rd_out : rd_out_ux_type;
	signal rd_out0_regged : std_logic_vector(b-1 downto 0);
	signal round_reg, round_reg2 : std_logic_vector(5-(UF-1) downto 0);
	
	signal mreg_sel, min_sel : std_logic_vector(mw-1 downto 0);
	signal grouped, dg : std_logic_vector(b-1 downto 0);
	signal rin, rout, rdreg : std_logic_vector(b-1 downto 0);
	signal hp, hm : std_logic_vector(b-1 downto 0);
	signal hp_or_iv : std_logic_vector(b-1 downto 0);
	signal c : c_array_type;
	
	-- id, hash		   
	constant id_size : integer := 16;
	signal id_in : std_logic_vector(ppl*id_size-1 downto 0);
	
	
	-- debug  						   
	-- signal crdo : std_logic_vector(b-1 downto 0);
	-- constant zeros : std_logic_vector(b-1-crw downto 0) := (others => '0');
	-- signal rout_m, rd_out_m,crd_out_m, rdreg_m, rin_m : std_logic_matrix;
	
begin	
	-- ///////////////////////////////////////////////////////
	-- FSM1 SECTION
	fsm1_section : for i in 0 to ppl-1 generate
		decounter_gen : decountern 
			generic map ( N => 16, sub => 1 ) 
			port map ( clk => clk, rst => '0', load => lc(i), en => ec(i), input => din(ppl*w-(i+1)*w+log2mw+15 downto ppl*w-(i+1)*w+log2mw), output => c(i));
		zc0(i) <= '1' when c(i) = 0 else '0';

        -- padding unit
        rem1_gen : 
            decountern 
            generic map ( N => 3, sub => 1 ) 
            port map ( clk => clk, rst => clr_lb(i), load => lc(i), en => en_lb(i), 
                      input => din(8 + (PPL-i-1)*W downto 6 + (PPL-i-1)*W), 
                      output => remaining_bits(i)(8 downto 6));	
        
        rem2_reg : 
            regn generic map ( N => 6, init => "000000" ) 
            port map ( clk => clk, rst => clr_lb(i), en => lc(i), 
                       input => din(5 + (PPL-i-1)*W downto 0 + (PPL-i-1)*W), 
                       output => remaining_bits(i)(5 downto 0));		
        
        sel_pad_location_lookup(i) <= lookup_sel_lvl2_64bit_pad(conv_integer(remaining_bits(i)(5 downto 3)));
        sel_pad_location(i) <= (others => '0') when spos(0 + (PPL-i-1)*2) = '0' else sel_pad_location_lookup(i);
        
        sel_din_lookup(i) <= lookup_sel_lvl1_64bit_pad(conv_integer(remaining_bits(i)(5 downto 3)));
        sel_din(i) <= (others => '1') when spos(1 + (PPL-i-1)*2) = '1' else sel_din_lookup(i);
        
        comp_rem_e0(i) <= '1' when remaining_bits(i) = 0 else '0';
        comp_lb_e0(i) <= '1' when remaining_bits(i)(8 downto 6) = "000" else '0';
        
		sel_pad_2bits(i) <= sel_pad((PPL-i)*2-1 downto (PPL-i-1)*2);
        pad_unit : entity work.jh_byte_pad(struct)
            port map (						
            din => din((PPL-i)*W-1 downto (PPL-i-1)*W), dout => din_padded(i), sel_pad => sel_pad_2bits(i), 
            sel_pad_location => sel_pad_location(i), sel_din => sel_din(i),
            clk => clk, rst => rst_size(i), en_size => en_size(i) );
            
		shfin_gen : sipo 
			generic map ( N => mw,M => w) 
			port map (clk => clk, en => ein(i), input => din_padded(i), output => min(i) );
		min_reg : regn 
			generic map ( N => mw, init => mwzeros ) 
			port map ( clk => clk, rst => '0', en => emsg(i), input => min(i), output => mreg(i) );
	end generate;

	-- ///////////////////////////////////////////////////////
	-- FSM2 SECTION
	-- Message selectors	 
	ppl2 : if ppl = 2 generate
		min_sel <= min(0) when smsg(0) = '0' else min(1);
		mreg_sel <= mreg(0) when smsg(0) = '0' else mreg(1);
	end generate;	   
	
	ppl4 : if ppl = 4 generate   
		with smsg(1 downto 0) select		
		min_sel <= min(0) when "00", min(1) when "01", min(2) when "10", min(3) when others;
		with smsg(1 downto 0) select		
		mreg_sel <= mreg(0) when "00", mreg(1) when "01", mreg(2) when "10", mreg(3) when others;
	end generate;	   
	
	-- input to r reg
	hm <= ( min_sel xor hp_or_iv(b-1 downto b/2) ) & hp_or_iv(b/2-1 downto 0);	
	hp_or_iv <= iv when sf = '1' else hp;	
	
	-- group (rearrange them into correct order)
	grouped <= form_group( hm, b, crw );			

	rin <= grouped when schain = '1' else rd_out(UF-1);

	-- Input Register to Round
	rreg_gen : regn 
		generic map ( N => b, init => bzeros ) 
		port map ( clk => clk, rst => '0', en => er, input => rin, output => rout );		
			
	--- RC	
	process ( clk )         
	begin
		if rising_edge( clk ) then
			crd_out <= crd_out_pre;	  			
		end if;
	end process;
	
	
	-- ROUND			
	UnrolledX1Gen: if UF = 1 generate
		crd_out_pre(0) <= rc_cons( conv_integer(round));
		rd_gen 	: entity work.jh_rd_ppl(struct) 
			generic map ( bw => b, cw => crw  ) 
			port map ( clk => clk, input => rout, cr => crd_out(0), output => rd_out(0) ); 
	end generate;
	UnrolledX2Gen: if UF = 2 generate					  
		crd_out_pre(0) <= rc_cons_half(0, conv_integer(round));
		
		
		stage0_reg : process ( clk ) 
		begin			   
			if rising_edge( clk ) then
				rd_out0_regged <= rd_out(0);  
				round_reg <= round;	 
				round_reg2 <= round_reg;
			end if;
		end process;
		
		ppl2 : if ppl = 2 generate 		  
			crd_out_pre(1) <= rc_cons_half(1, conv_integer(round_reg));
			rd_gen1 	: entity work.jh_rd(struct) 
				generic map ( bw => b, cw => crw  ) 
				port map ( input => rout, cr => crd_out(0), output => rd_out(0) ); 	
			rd_gen2 	: entity work.jh_rd(struct) 
				generic map ( bw => b, cw => crw  ) 
				port map ( input => rd_out0_regged, cr => crd_out(1), output => rd_out(1) ); 
		end generate;
		ppl4 : if ppl = 4 generate	  
			crd_out_pre(1) <= rc_cons_half(1, conv_integer(round_reg2));
			rd_gen1 	: entity work.jh_rd_ppl(struct) 
				generic map ( bw => b, cw => crw  ) 
				port map ( clk => clk, input => rout, cr => crd_out(0), output => rd_out(0) ); 
			rd_gen2 	: entity work.jh_rd_ppl(struct) 
				generic map ( bw => b, cw => crw  ) 
				port map ( clk => clk, input => rd_out0_regged, cr => crd_out(1), output => rd_out(1) ); 
		end generate;			
	end generate;			
		
	rdreg_gen : regn 
		generic map ( N => b, init => bzeros ) 
		port map ( clk => clk, rst => '0', en => er, input => rd_out(UF-1), output => rdreg );		
	
	dg <= degroup( rdreg, b, crw );
 	hp <= dg(b-1 downto b/2) & ( mreg_sel xor dg(b/2-1 downto 0) );	 
	-- ///////////////////////////////////////////////////////
	-- FSM3 SECTION
	-- Round Finalization / Load new block / Output generation
	fsm3_output : entity work.ppl_output_unit(struct) 
		generic map ( h => HS, w => w, ppl => ppl ) 
		port map ( clk => clk, rst => rst,
					hash_result => hp(HS-1 downto 0), id_in => id_in,
					sid => sid, shift_dout => shift_dout, load_hash => load_hash, 
					load_id => load_id, sout => sout,
					dout => dout );			 
			
	id_in_gen : for i in 0 to ppl-1 generate
		id_in((ppl-i)*id_size-1 downto (ppl-i-1)*id_size) <= din(ppl*w-(i+1)*w+15 downto ppl*w-(i+1)*w);
	end generate; 		

--	--debug
	 --rin_m <= blk2wordmatrix_inv( rin );
--	  rout_m <= blk2wordmatrix_inv( rout );
--	  rd_out_m <= blk2wordmatrix_inv( rd_out(0) );
--	  rdreg_m <= blk2wordmatrix_inv( rdreg );
--	dg_m <= blk2wordmatrix_inv( dg );
--	hp_m <= blk2wordmatrix_inv( hp );
--	hm_m <= blk2wordmatrix_inv( hm );
--	crdo <= crd_out(ux-1) & zeros;
--	crd_out_m <= blk2wordmatrix_inv( crdo );
end struct;		   		   