-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use work.sha3_pkg.all;
use work.sha3_jh_package.all;

entity jh_fsm2_mem is 
	generic (
		ppl : integer := 2;
		UF : integer := 1
	);
	port (
		-- global
		clk : in std_logic;
		rst : in std_logic;
		
		-- datapath	  
		smsg : out std_logic_vector(log2(ppl)-1 downto 0);	
		emsg, load_hash : out std_logic_vector(ppl-1 downto 0);
		schain : out std_logic;			
		er : out std_logic; 		
		sf : out std_logic;
		round : out std_logic_vector(5-(UF-1) downto 0);
		-- control				   
			--fsm1 hand shake signals
		block_ready_clr : out std_logic_vector(ppl-1 downto 0);		
		msg_end_clr 	: out std_logic_vector(ppl-1 downto 0);
		
		block_ready		: in std_logic_vector(ppl-1 downto 0);
		msg_end 		: in std_logic_vector(ppl-1 downto 0);	
				
			--fsm3 handshake signals
		output_busy_set  : out std_logic_vector(ppl-1 downto 0);
		output_busy		 : in  std_logic_vector(ppl-1 downto 0)
	);							   
end jh_fsm2_mem;

architecture beh of jh_fsm2_mem is			  
	constant roundnr 		: integer := 42/UF + 1;  -- Add an additional round for block finalization
	constant log2roundnr	: integer := log2( roundnr );
	constant log2roundnrzeros : std_logic_vector(log2roundnr-1 downto 0) := (others => '0') ;
		
	signal pc : std_logic_vector(log2roundnr-1 downto 0);
	signal ziroundnr,zi0, zi15, ei, li : std_logic;
	signal output_data_s : std_logic_vector(ppl-1 downto 0);
	
	signal msg_no, ppl_stage_cnt : std_logic_vector(log2(ppl)-1 downto 0);
	signal valid_set, valid_clr, valid : std_logic_vector(ppl-1 downto 0);
	signal sf_set, sf_clr, sf_sig : std_logic_vector(ppl-1 downto 0);
	signal schain_sig: std_logic;
	signal block_ready_true: std_logic_vector(PPL-1 downto 0);
	
	constant log2pplzeros : std_logic_vector(log2(ppl)-1 downto 0) := (others =>'0');
	constant log2pplones : std_logic_vector(log2(ppl)-1 downto 0) := (others =>'1');
begin	
	-- Internals
	proc_counter_gen : countern generic map ( N => log2roundnr ) port map ( clk => clk, rst => rst, load => li, en => ei, input => log2roundnrzeros, output => pc);
	
	pplstage_cnt: process ( clk )
	begin
		if rising_edge( clk ) then
		
			if rst = '1' or ei = '1' or li = '1' then
				ppl_stage_cnt <= (others => '0');
			else
				ppl_stage_cnt <= ppl_stage_cnt + 1;
			end if;
		end if;
	end process;
	
	ziroundnr <= '1' when pc = conv_std_logic_vector(roundnr-1,log2roundnr) else '0';			 
	zi0 <= '1' when pc = conv_std_logic_vector(0,log2roundnr) else '0';
	zi15 <= '1' when pc = conv_std_logic_vector(15,log2roundnr) else '0';
	round <= pc;
		
	
	li <= '1' when (ziroundnr = '1' and ppl_stage_cnt = log2pplzeros) else '0';
	-- If stall is need, modify 'ei' and 'msg_no' to stall
	ei <= '1' when (ppl_stage_cnt = log2pplones) else '0';
	
	-- msg no counter
	mno_cnt: process( clk ) 
	begin
		if rising_edge( clk ) then
			if rst = '1' or ei = '1' then
				msg_no <= (others => '0');
			else	  				
				msg_no <= msg_no + 1;
			end if;
		end if;
	end process;   
	smsg <= msg_no;
	
	
	---- output logic	   					
	er 		<= '1';						 
	
	gg1: for i in 0 to ppl-1 generate	 		
		emsg(i) <= '1' when (schain_sig = '1' and msg_no = conv_std_logic_vector(i,log2(ppl)) and block_ready_true(i) = '1') else '0';
		output_data_s(i) <= '1' when (schain_sig = '1' and msg_end(i) = '1' and output_busy(i) = '0' and msg_no = conv_std_logic_vector(i,log2(ppl)) and valid(i) = '1') else '0';
		output_busy_set(i) 	<= output_data_s(i);
		load_hash(i)		<= output_data_s(i);	   
		msg_end_clr(i)		<= output_data_s(i);	
	   	
			   
		sf_set(i) <= '1' when (zi15 = '1' and msg_end(i)  = '1')  else '0';

		sf_gen : sr_reg 
			generic map ( init => '1' )
			port map ( rst => rst, clk => clk, set => sf_set(i), clr => sf_clr(i), output => sf_sig(i));		
		
		valid_gen : sr_reg 
			generic map ( init => '0' )
			port map ( rst => rst, clk => clk, set => valid_set(i), clr => valid_clr(i), output => valid(i));		
	end generate;	
	
	Ppl2Gen:
		if PPL = 2 generate
			sf <= sf_sig(0) when msg_no(0) = '0' else sf_sig(1); 
		end generate;
	Ppl4Gen:
		if PPL = 4 generate
			with msg_no(1 downto 0) select
			sf <= sf_sig(0) when "00", sf_sig(1) when "01", sf_sig(2) when "10", sf_sig(3) when others;
		end generate;	
						   
	sf_clr(0) <= '1' when (ziroundnr = '1' and block_ready_true(0) = '1' ) else '0'; 
	valid_set(0) <= '1' when (ziroundnr = '1' and block_ready_true(0) = '1') else '0'; 
	valid_clr(0) <= '1' when (ziroundnr = '1' and block_ready_true(0) = '0') else '0'; 	
	gg2 : for i in 1 to ppl-1 generate			
		sf_clr(i) <= '1' when (zi0 = '1' and ppl_stage_cnt = conv_std_logic_vector(i-1,log2(ppl)) and block_ready_true(i) = '1') else '0'; 
		valid_set(i) <= '1' when (zi0 = '1' and ppl_stage_cnt = conv_std_logic_vector(i-1,log2(ppl)) and block_ready_true(i) = '1') else '0'; 
		valid_clr(i) <= '1' when (zi0 = '1' and ppl_stage_cnt = conv_std_logic_vector(i-1,log2(ppl)) and block_ready_true(i) = '0') else '0'; 
	end generate;
	block_ready_clr <= valid_set;
																					   	
	schain_sig <= '1' when (ziroundnr = '1' or (pc = 0 and ppl_stage_cnt /= ppl-1)) else '0';
	schain <= schain_sig;
	
	BlockReadyTrueReg : process ( clk ) 	-- Allow everything to be synchronize to round number
	begin
		if rising_edge( clk ) then	 
			if rst = '1' then
				block_ready_true <= (others => '0');				
			elsif (pc = roundnr-2 and ppl_stage_cnt = ppl-1) then  				
				block_ready_true <= block_ready;
			end if;
		end if;
	end process;
	
end beh;
			  